'\"
'\" Generated from file 'doc/scrobj.tcl' by tcllib/doctools with format 'nroff'
'\"
'\" -*- tcl -*- doctools manpage $Id: scrobj.n,v 1.6 2008/12/22 10:55:00 cn Exp $
.TH "scrobj" n 1.0  "scripted Tcl_Objects"
.BS
.SH "NAME"
scrobj \- scripted Tcl_Objects
.SH "SYNOPSIS"
package require \fBTcl  8.5\fR
.sp
package require \fBscrobj  1.0\fR
.sp
\fBscrobj\fR \fBregister\fR \fItypeName\fR \fIupdateCode\fR \fIparseCode\fR\fR
.sp
\fBscrobj\fR \fBconvert\fR \fItypeName\fR \fIstring\fR ?extraArg?\fR
.sp
\fBscrobj\fR \fBvalue\fR \fItypeName\fR \fIintRep\fR ?extraArg?\fR
.sp
\fBscrobj\fR \fBeval\fR \fItypeName\fR \fIcode\fR\fR
.sp
\fBscrobj\fR \fBinfo\fR ?typeName?\fR
.sp
.BE
.SH "DESCRIPTION"
.PP
Tcl's fundamental mantra says that "everything is a string" and
much of Tcl's simplicity derives from this rule. Strings can,
however, carry with them an internal representation as some other
data type, and this is one reason for Tcl's astonishing performance.
For example, Tcl maintains such an internal representation for integers,
lists, bytecodes, and many other types.
.PP
Extension writers have long been able to add new custom types
using Tcl's C interface. The scrobj package allows you to do the same
from the script level.
.PP
Unfortunately, in order to achieve this the package has to
violate a basic assumption of Tcl's bytecode engine, namely that
it is not getting invoked recursively from a call to Tcl_GetString.
A consequence is that scrobj can be used to crash the application,
although this does require a rather contrived setup.
.PP
To prevent such crashes in the legitimate cases the package uses
a dedicated Tcl interpreter for every registered type. Both
Tcl 8.5 and Tcl 8.6 seem to tolerate recursive invocations
of the bytecode engine, as long as they execute in different
interpreters.
.PP
.SH "Usage"
.TP
\fBscrobj\fR \fBregister\fR \fItypeName\fR \fIupdateCode\fR \fIparseCode\fR\fR
To register a new data type you have to write code
to transform the string value to the internal representation
and vice versa. This code then needs to be registered using
\fBscrobj\fR \fBregister\fR.
.sp
.nf
scrobj register mytype {
    intRep {
	# code to generate the string
	# from the internal representation
	set dummy "<$intRep>"
    }
} {
    stringVal {
	# code to generate the internal
	# representation from the string
	regexp -- {^<([0-9]+)>$} $stringVal -> irp
	set irp
    }
}
.fi
.sp
The call to \fBscrobj\fR \fBregister\fR implicitly creates a new Tcl
interpreter where the conversion routines will be executed.
.TP
\fBscrobj\fR \fBconvert\fR \fItypeName\fR \fIstring\fR ?extraArg?\fR
Return the internal representation corresponding to \fIstring\fR
for the type \fItypeName\fR. If the internal representation is
already available, that value will be returned directly. Otherwise
the internal representation will first be generated by applying
the type's \fIparseCode\fR to the given \fIstring\fR.
.sp
.nf
# convert "<4711>" and return the internal representation "4711"
scrobj convert mytype <4711>
.fi
.TP
\fBscrobj\fR \fBvalue\fR \fItypeName\fR \fIintRep\fR ?extraArg?\fR
Generate a value from the internal representation.
.sp
.nf
# generate a value for type mytype
set a [scrobj value mytype 123]

# the next call will \fInot\fR invoke the
# parseCode because $a already has an internal
# representation for type "mytype"
set b [scrobj convert mytype $a]

# the next line will run the updateCode for mytype
set c "a is $a" ;# -> "a is <123>"
.fi
.TP
\fBscrobj\fR \fBeval\fR \fItypeName\fR \fIcode\fR\fR
Execute some \fIcode\fR in the interpreter for type \fItypeName\fR.
.TP
\fBscrobj\fR \fBinfo\fR ?typeName?\fR
Return the list of all registered types, or information about
one specific type.
.SH "Parametrized types"
The \fBconvert\fR and \fBvalue\fR subcommands allow to
specify an optional extra argument. This can be used
to parametrize a data type by a value:
.TP
Example:
.nf
# register a parametrized type that uses a configurable
# regular expression for the conversion
scrobj register rexp {
    irep {
        # string is kept in internal rep
        lindex $irep 1
    }
} {
    {inString regExp} {
        regexp -- $regExp $inString -> res
        list $res $inString
    }
}

set pat1 {^foo([0-9]+)$}
set pat2 {^foo([0-9]+)$}
set pat3 {^fo(o[0-9]+)$}

set a [scrobj value rexp [scrobj convert rexp foo17 $pat1] $pat1]

# The next line will use the internal representation
# of $a that's already there
set b [scrobj convert rexp $a $pat1]

# The next line will re-generate the internal
# representation, because $pat1 and $pat2
# are not represented by the same Tcl_Object.
set c [scrobj convert rexp $a $pat2]

# The next line also re-generates the internal
# representation. This time the return value
# is actually different.
set d [scrobj convert rexp $a $pat3]
.fi
.SH "Application examples"
scrobj comes equipped with two packages,
\fIscrobj::expression\fR and \fIscrobj::rational\fR,
which illustrate the use of scrobj types.
.TP
scrobj::expression
\fIscrobj::expression\fR implements a parser for simple arithmetic
expressions.
.sp
.nf
package require scrobj::expression 1.0

scrobj convert scrobj::expression {23*foo(bar(12),3+$a/7)} {
    opns ::operatorns::
    funcns ::funcns::
}

# the result of the conversion is Tcl code that
# implements the computation:
{
    upvar 1 a 3
    set 1 [::funcns::bar [eval {set dummy 12}]]
    set 2 [::funcns::foo [eval {set 1}] [eval {
	set 4 [::operatorns::/ $3 7]
	set 5 [::operatorns::+ 3 $4]
	set 5
    }]]
    set 6 [::operatorns::* 23 $2]
    set 6
} {23*foo(bar(12),3+$a/7)}
.fi
.sp
The parser is implemented in Tcl using regular expressions.
It doesn't have to be fast, because the script
will be cached in the internal representation once it is
generated.
.TP
scrobj::rational
The \fIscrobj::rational\fR package implements rational number
arithmetic. A rational number p/q is represented internally as
the list \fB[list $p $q]\fR.
.sp
.nf
package require scrobj::rational 1.0

scrobj::rational numerator 17272/9128 ;# -> 17272
scrobj::rational + 1727/162 8348/1731 ;# -> 1447271/93474
.fi
.sp
\fIscrobj::rational\fR uses \fIscrobj::expression\fR to implement
its own version of the \fBexpr\fR command:
.sp
.nf
package require scrobj::rational 1.0

# register some custom functions
namespace eval rational {
    namespace eval func {
	proc f1 {a b} {
	    ::scrobj::rational expression {3*$a-17/93*$b}
	}
	proc f2 {x} {
	    ::scrobj::rational expression {2*$x-1}
	}
    }
}

set a 72/841

scrobj::rational expression {15*f1($a,3-$a) - f2(17/18+$a)/9}
# -> -8988239/2111751
.fi
.SH "KEYWORDS"
Tcl_Object, expression, rational, scrobj
